# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc helper functions."""

import struct

from absl.testing import absltest
from absl.testing import parameterized
import mock
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import tensor_utils


class PackTensorTests(parameterized.TestCase):

  @parameterized.parameters(
      (np.float32(2.5), 'floats'),
      (2.5, 'doubles'),
      (np.int32(-25), 'int32s'),
      (np.int64(-25), 'int64s'),
      (np.frombuffer(b'\xF0\xF1\xF2\xF3', np.uint32)[0], 'uint32s'),
      (np.frombuffer(b'\xF0\xF1\xF2\xF3\xF4\xF5\xF6\xF7',
                     np.uint64)[0], 'uint64s'),
      (True, 'bools'),
      (False, 'bools'),
      ('foo', 'strings'),
  )
  def test_pack_scalars(self, scalar, expected_payload):
    tensor = tensor_utils.pack_tensor(scalar)
    self.assertEqual([], tensor.shape)
    self.assertEqual([scalar], getattr(tensor, expected_payload).array)

  @parameterized.parameters(
      (np.int8(-25), 'b', 'int8s'),
      (np.uint8(250), 'B', 'uint8s'),
  )
  def test_pack_scalar_bytes(self, scalar, fmt, expected_payload):
    tensor = tensor_utils.pack_tensor(scalar)
    self.assertEqual([], tensor.shape)
    actual = struct.unpack(fmt, getattr(tensor, expected_payload).array)
    self.assertEqual(scalar, actual)

  @parameterized.parameters(
      (25, np.float32, 'floats'),
      (25, np.float64, 'doubles'),
      (25, np.int32, 'int32s'),
      (25, np.int64, 'int64s'),
      (25, np.uint32, 'uint32s'),
      (25, np.uint64, 'uint64s'),
      (True, np.bool, 'bools'),
      (False, np.bool, 'bools'),
      ('foo', np.str, 'strings'),
  )
  def test_pack_scalars_specific_dtype(self, scalar, dtype, expected_payload):
    tensor = tensor_utils.pack_tensor(scalar, dtype)
    self.assertEqual([], tensor.shape)
    self.assertEqual([scalar], getattr(tensor, expected_payload).array)

  def test_pack_with_dm_env_rpc_data_type(self):
    tensor = tensor_utils.pack_tensor([5], dm_env_rpc_pb2.DataType.FLOAT)
    self.assertEqual([5], tensor.floats.array)

  @parameterized.parameters(
      ([np.int8(-25), np.int8(-23)], '2b', 'int8s'),
      ([np.uint8(249), np.uint8(250)], '2B', 'uint8s'),
  )
  def test_pack_bytes_array(self, scalar, fmt, expected_payload):
    tensor = tensor_utils.pack_tensor(scalar)
    self.assertEqual([2], tensor.shape)
    actual = struct.unpack(fmt, getattr(tensor, expected_payload).array)
    np.testing.assert_array_equal(scalar, actual)

  @parameterized.parameters(
      (np.array([1.0, 2.0], dtype=np.float32), 'floats'),
      (np.array([1.0, 2.0], dtype=np.float64), 'doubles'),
      ([1.0, 2.0], 'doubles'),
      (np.array([1, 2], dtype=np.int32), 'int32s'),
      (np.array([1, 2], dtype=np.int64), 'int64s'),
      (np.array([1, 2], dtype=np.uint32), 'uint32s'),
      (np.array([1, 2], dtype=np.uint64), 'uint64s'),
      ([True, False], 'bools'),
      (np.array([True, False]), 'bools'),
      (['foo', 'bar'], 'strings'),
  )
  def test_pack_arrays(self, array, expected_payload):
    tensor = tensor_utils.pack_tensor(array)
    self.assertEqual([2], tensor.shape)
    packed_array = getattr(tensor, expected_payload).array
    np.testing.assert_array_equal(array, packed_array)

  def test_packed_rowmajor(self):
    array2d = np.array([[1, 2], [3, 4], [5, 6]], dtype=np.int32)
    tensor = tensor_utils.pack_tensor(array2d)
    self.assertEqual([3, 2], tensor.shape)
    np.testing.assert_array_equal([1, 2, 3, 4, 5, 6], tensor.int32s.array)

  def test_mixed_scalar_types_raises_exception(self):
    with self.assertRaises(TypeError):
      tensor_utils.pack_tensor(['hello!', 75], dtype=np.float32)

  def test_jagged_arrays_throw_exceptions(self):
    with self.assertRaises(ValueError):
      tensor_utils.pack_tensor([[1, 2], [3, 4, 5]])

  def test_class_instance_throw_exception(self):

    class Foo(object):
      pass

    with self.assertRaises(ValueError):
      tensor_utils.pack_tensor(Foo())

  def test_compress_integers_to_1_element_when_all_same(self):
    array = np.array([1, 1, 1, 1, 1, 1], dtype=np.uint32)
    packed = tensor_utils.pack_tensor(array, try_compress=True)
    self.assertEqual([6], packed.shape)
    self.assertEqual([1], packed.uint32s.array)

  def test_compress_floats_to_1_element_when_all_same(self):
    array = np.array([1.5, 1.5, 1.5, 1.5, 1.5, 1.5], dtype=np.float32)
    packed = tensor_utils.pack_tensor(array, try_compress=True)
    self.assertEqual([6], packed.shape)
    self.assertEqual([1.5], packed.floats.array)

  def test_compress_strings_to_1_element_when_all_same(self):
    array = np.array(['foo', 'foo', 'foo', 'foo'], dtype=np.str_)
    packed = tensor_utils.pack_tensor(array, try_compress=True)
    self.assertEqual([4], packed.shape)
    self.assertEqual(['foo'], packed.strings.array)

  def test_compress_multidimensional_arrays_to_1_element_when_all_same(self):
    array = np.array([[4, 4], [4, 4]], dtype=np.int32)
    packed = tensor_utils.pack_tensor(array, try_compress=True)
    self.assertEqual([2, 2], packed.shape)
    self.assertEqual([4], packed.int32s.array)

  def test_doesnt_compress_if_not_asked_to(self):
    array = np.array([1, 1, 1, 1, 1, 1], dtype=np.uint32)
    packed = tensor_utils.pack_tensor(array)
    self.assertEqual([6], packed.shape)
    self.assertEqual([1, 1, 1, 1, 1, 1], packed.uint32s.array)

  def test_ask_to_compress_but_cant(self):
    array = np.array([1, 1, 2, 1, 1, 1], dtype=np.uint32)
    packed = tensor_utils.pack_tensor(array, try_compress=True)
    self.assertEqual([6], packed.shape)
    self.assertEqual([1, 1, 2, 1, 1, 1], packed.uint32s.array)


class UnpackTensorTests(parameterized.TestCase):

  @parameterized.parameters(
      np.float32(2.5),
      np.float64(2.5),
      np.int8(-25),
      np.int32(-25),
      np.int64(-25),
      np.uint8(250),
      np.frombuffer(b'\xF0\xF1\xF2\xF3', np.uint32)[0],
      np.frombuffer(b'\xF0\xF1\xF2\xF3\xF4\xF5\xF6\xF7', np.uint64)[0],
      True,
      False,
      'foo',
  )
  def test_unpack_scalars(self, scalar):
    tensor = tensor_utils.pack_tensor(scalar)
    round_trip = tensor_utils.unpack_tensor(tensor)
    self.assertEqual(scalar, round_trip)

  @parameterized.parameters(
      ([np.float32(2.5), np.float32(3.5)],),
      ([np.float64(2.5), np.float64(3.5)],),
      ([np.int8(-25), np.int8(-23)],),
      ([np.int32(-25), np.int32(-23)],),
      ([np.int64(-25), np.int64(-23)],),
      ([np.uint8(250), np.uint8(249)],),
      ([np.uint32(1), np.uint32(2)],),
      ([np.uint64(1), np.uint64(2)],),
      ([True, False],),
      (['foo', 'bar'],),
  )
  def test_unpack_arrays(self, array):
    tensor = tensor_utils.pack_tensor(array)
    round_trip = tensor_utils.unpack_tensor(tensor)
    np.testing.assert_array_equal(array, round_trip)

  def test_unpack_multidimensional_arrays(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.floats.array[:] = [1, 2, 3, 4, 5, 6, 7, 8]
    tensor.shape[:] = [2, 4]
    round_trip = tensor_utils.unpack_tensor(tensor)
    expected = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
    np.testing.assert_array_equal(expected, round_trip)

  def test_too_few_elements(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.floats.array[:] = [1, 2, 3, 4]
    tensor.shape[:] = [2, 4]
    with self.assertRaisesRegexp(ValueError, 'cannot reshape array'):
      tensor_utils.unpack_tensor(tensor)

  def test_too_many_elements(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.floats.array[:] = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    tensor.shape[:] = [2, 4]
    with self.assertRaisesRegexp(ValueError, 'cannot reshape array'):
      tensor_utils.unpack_tensor(tensor)

  def test_float_broadcasts_1_element_to_all_elements(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.floats.array[:] = [1]
    tensor.shape[:] = [4]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array([1, 1, 1, 1], dtype=np.float32)
    np.testing.assert_array_equal(expected, unpacked)

  def test_integer_broadcasts_1_element_to_all_elements(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.int32s.array[:] = [1]
    tensor.shape[:] = [4]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array([1, 1, 1, 1], dtype=np.int32)
    np.testing.assert_array_equal(expected, unpacked)

  def test_unsigned_integer_broadcasts_1_element_to_all_elements(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor_utils._BytesWrapper(tensor.uint8s, signed=False)[:] = [1]
    tensor.shape[:] = [4]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array([1, 1, 1, 1], dtype=np.uint8)
    np.testing.assert_array_equal(expected, unpacked)

  def test_string_broadcasts_1_element_to_all_elements(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.strings.array[:] = ['foo']
    tensor.shape[:] = [4]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array(['foo', 'foo', 'foo', 'foo'], dtype=np.str_)
    np.testing.assert_array_equal(expected, unpacked)

  def test_broadcasts_to_multidimensional_arrays(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.int32s.array[:] = [4]
    tensor.shape[:] = [2, 2]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array([[4, 4], [4, 4]], dtype=np.int32)
    np.testing.assert_array_equal(expected, unpacked)

  def test_negative_dimension(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.int32s.array[:] = [1, 2, 3, 4]
    tensor.shape[:] = [-1]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array([1, 2, 3, 4], dtype=np.int32)
    np.testing.assert_array_equal(expected, unpacked)

  def test_negative_dimension_in_matrix(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.int32s.array[:] = [1, 2, 3, 4, 5, 6]
    tensor.shape[:] = [2, -1]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.int32)
    np.testing.assert_array_equal(expected, unpacked)

  def test_two_negative_dimensions_in_matrix(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.int32s.array[:] = [1, 2, 3, 4, 5, 6]
    tensor.shape[:] = [-1, -2]
    with self.assertRaisesRegexp(ValueError, 'one unknown dimension'):
      tensor_utils.unpack_tensor(tensor)

  def test_negative_dimension_single_element(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.int32s.array[:] = [1]
    tensor.shape[:] = [-1]
    unpacked = tensor_utils.unpack_tensor(tensor)
    expected = np.array([1], dtype=np.int32)
    np.testing.assert_array_equal(expected, unpacked)

  def test_unknown_type_raises_error(self):
    tensor = mock.MagicMock()
    tensor.WhichOneof.return_value = 'foo'
    with self.assertRaisesRegexp(TypeError, 'type foo'):
      tensor_utils.unpack_tensor(tensor)

  def test_scalar_with_too_many_elements_raises_error(self):
    tensor = dm_env_rpc_pb2.Tensor()
    tensor.int32s.array[:] = [1, 2, 3]
    with self.assertRaisesRegexp(ValueError, '3 element'):
      tensor_utils.unpack_tensor(tensor)


class GetTensorTypeTests(absltest.TestCase):

  def test_float(self):
    tensor = tensor_utils.pack_tensor(1.25)
    self.assertEqual(np.float64, tensor_utils.get_tensor_type(tensor))

  def test_unknown_tensor_type(self):
    mock_tensor = mock.MagicMock()
    mock_tensor.WhichOneof.return_value = 'foo'
    with self.assertRaisesRegexp(TypeError, 'foo'):
      tensor_utils.get_tensor_type(mock_tensor)


class GetTensorSpecTypeTests(absltest.TestCase):

  def test_float(self):
    self.assertEqual(
        np.float32,
        tensor_utils.data_type_to_np_type(dm_env_rpc_pb2.DataType.FLOAT))

  def test_proto_type(self):
    with self.assertRaises(TypeError):
      tensor_utils.data_type_to_np_type(dm_env_rpc_pb2.DataType.PROTO)

  def test_unknown_type(self):
    with self.assertRaises(TypeError):
      tensor_utils.data_type_to_np_type(30)


class BytesWrapperTests(absltest.TestCase):

  def test_unsupported_indexing_on_write_raises_error(self):
    tensor = dm_env_rpc_pb2.Tensor()
    wrapper = tensor_utils._BytesWrapper(tensor.uint8s, signed=False)
    with self.assertRaisesRegexp(ValueError, 'index'):
      wrapper[0] = 0


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Manager class to manage the dm_env_rpc UID system."""

import numpy as np

from dm_env_rpc.v1 import tensor_utils


def _assert_shapes_match(tensor, dm_env_rpc_spec):
  """Raises ValueError if shape of tensor and spec don't match."""
  if not np.array_equal(tensor.shape, dm_env_rpc_spec.shape):
    raise ValueError(
        'Received dm_env_rpc tensor {} with shape {} but spec has shape {}.'
        .format(dm_env_rpc_spec.name, tensor.shape, dm_env_rpc_spec.shape))


class SpecManager(object):
  """Manages transitions between Python dicts and dm_env_rpc UIDs.

  To make sending and receiving actions and observations easier for dm_env_rpc,
  this helps manage the transition between UID-keyed dicts mapping to dm_env_rpc
  tensors and string-keyed dicts mapping to scalars, lists, or NumPy arrays.
  """

  def __init__(self, specs):
    """Builds the SpecManager from the given dm_env_rpc specs.

    Args:
      specs: A dict mapping UIDs to dm_env_rpc TensorSpecs, similar to what is
        stored in `actions` and `observations` in ActionObservationSpecs.
    """
    self._name_to_uid = {
        spec.name: uid for uid, spec in specs.items()
    }
    self._uid_to_name = {
        uid: spec.name for uid, spec in specs.items()
    }
    if len(self._name_to_uid) != len(self._uid_to_name):
      raise ValueError('There are duplicate names in the tensor specs.')

    self._specs_by_uid = specs
    self._specs_by_name = {
        spec.name: spec for spec in specs.values()
    }

  def name_to_uid(self, name):
    """Returns the UID for the given name."""
    return self._name_to_uid[name]

  def uid_to_name(self, uid):
    """Returns the name for the given UID."""
    return self._uid_to_name[uid]

  def name_to_spec(self, name):
    """Returns the dm_env_rpc TensorSpec named `name`."""
    return self._specs_by_name[name]

  def uid_to_spec(self, uid):
    """Returns the dm_env_rpc TensorSpec for the given UID."""
    return self._specs_by_uid[uid]

  def names(self):
    """Returns the spec names in no particular order."""
    return self._name_to_uid.keys()

  def uids(self):
    """Returns the spec UIDs in no particular order."""
    return self._uid_to_name.keys()

  def unpack(self, dm_env_rpc_tensors):
    """Unpacks a dm_env_rpc uid-to-tensor map to a name-keyed Python dict.

    Args:
      dm_env_rpc_tensors: A dict mapping UIDs to dm_env_rpc tensor protos.

    Returns:
      A dict mapping names to scalars and arrays.
    """
    unpacked = {}
    for uid, tensor in dm_env_rpc_tensors.items():
      name = self._uid_to_name[uid]
      dm_env_rpc_spec = self.name_to_spec(name)
      _assert_shapes_match(tensor, dm_env_rpc_spec)
      tensor_dtype = tensor_utils.get_tensor_type(tensor)
      spec_dtype = tensor_utils.data_type_to_np_type(dm_env_rpc_spec.dtype)
      if tensor_dtype != spec_dtype:
        raise ValueError(
            'Received dm_env_rpc tensor {} with dtype {} but spec has dtype {}.'
            .format(name, tensor_dtype, spec_dtype))
      tensor_unpacked = tensor_utils.unpack_tensor(tensor)
      unpacked[name] = tensor_unpacked
    return unpacked

  def pack(self, tensors):
    """Packs a name-keyed Python dict to a dm_env_rpc uid-to-tensor map.

    Args:
      tensors: A dict mapping string names to scalars and arrays.

    Returns:
      A dict mapping UIDs to dm_env_rpc tensor protos.
    """
    packed = {}
    for name, value in tensors.items():
      dm_env_rpc_spec = self.name_to_spec(name)
      tensor = tensor_utils.pack_tensor(value, dtype=dm_env_rpc_spec.dtype)
      _assert_shapes_match(tensor, dm_env_rpc_spec)
      packed[self.name_to_uid(name)] = tensor
    return packed



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc helper functions."""

from absl.testing import absltest

import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import tensor_spec_utils
from dm_env_rpc.v1 import tensor_utils


class NpRangeInfoTests(absltest.TestCase):

  def test_floating(self):
    expected_min = np.finfo(np.float32).min
    actual_min = tensor_spec_utils._np_range_info(np.float32).min
    self.assertEqual(expected_min, actual_min)

  def test_integer(self):
    actual_min = tensor_spec_utils._np_range_info(np.uint32).min
    self.assertEqual(0, actual_min)

  def test_string_gives_error(self):
    with self.assertRaisesRegex(ValueError, 'numpy.str_'):
      _ = tensor_spec_utils._np_range_info(np.str_).min


class BoundsTests(absltest.TestCase):

  def test_unbounded_unsigned(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((0, 2**32-1), bounds)

  def test_unbounded_signed(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((-2**31, 2**31-1), bounds)

  def test_min_n_shape(self):
    minimum = np.array([[1, 2], [3, 4]])

    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = minimum.flatten().data.tolist()
    tensor_spec.shape[:] = minimum.shape
    bounds = tensor_spec_utils.bounds(tensor_spec)
    np.testing.assert_array_equal(minimum, bounds.min)
    np.testing.assert_array_equal(np.full(minimum.shape, 2**32 - 1), bounds.max)

  def test_max_n_shape(self):
    maximum = np.array([[1, 2], [3, 4]])

    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = maximum.flatten().data.tolist()
    tensor_spec.shape[:] = maximum.shape
    bounds = tensor_spec_utils.bounds(tensor_spec)
    np.testing.assert_array_equal(np.full(maximum.shape, 0), bounds.min)
    np.testing.assert_array_equal(maximum, bounds.max)

  def test_invalid_min_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1, 2]
    with self.assertRaisesRegex(ValueError,
                                'Scalar tensors must have exactly 1 element.*'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_invalid_max_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = [1, 2]
    tensor_spec.shape[:] = (2, 2)
    with self.assertRaisesRegex(ValueError,
                                'cannot reshape array of size .* into shape.*'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_min(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1]
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((1, 2**32-1), bounds)

  def test_max(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = [1]
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((0, 1), bounds)

  def test_min_and_max(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.min.int32s.array[:] = [-1]
    tensor_spec.max.int32s.array[:] = [1]
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((-1, 1), bounds)

  def test_broadcast_var_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.min.int32s.array[:] = [-1]
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.shape[:] = (-1,)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((-1, 1), bounds)

  def test_invalid_min_var_shape(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.min.int32s.array[:] = [-1, -1]
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.shape[:] = (-1,)
    with self.assertRaisesRegex(
        ValueError, "TensorSpec's with variable length shapes "
        'can only have scalar ranges.'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_min_broadcast(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1]
    tensor_spec.shape[:] = (2, 2)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    np.testing.assert_array_equal(np.full(tensor_spec.shape, 1), bounds.min)
    np.testing.assert_array_equal(
        np.full(tensor_spec.shape, 2**32 - 1), bounds.max)

  def test_max_broadcast(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32s.array[:] = [1]
    tensor_spec.shape[:] = (2, 2)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    np.testing.assert_array_equal(np.full(tensor_spec.shape, 0), bounds.min)
    np.testing.assert_array_equal(np.full(tensor_spec.shape, 1), bounds.max)

  def test_min_max_broadcast(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32s.array[:] = [1]
    tensor_spec.max.uint32s.array[:] = [2]
    tensor_spec.shape[:] = (4,)
    bounds = tensor_spec_utils.bounds(tensor_spec)
    np.testing.assert_array_equal(np.full(tensor_spec.shape, 1), bounds.min)
    np.testing.assert_array_equal(np.full(tensor_spec.shape, 2), bounds.max)

  def test_min_mismatches_type_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*uint32.*min.*int32'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_max_mismatches_type_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*uint32.*max.*int32'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_nonnumeric_type_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.STRING
    tensor_spec.max.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*non-numeric.*string'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_max_0_stays_0(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
    tensor_utils._BytesWrapper(tensor_spec.max.int8s, signed=False)[:] = [0]
    tensor_spec.name = 'foo'
    self.assertEqual((-128, 0), tensor_spec_utils.bounds(tensor_spec))

  def test_min_0_stays_0(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
    tensor_utils._BytesWrapper(tensor_spec.min.int8s, signed=False)[:] = [0]
    tensor_spec.name = 'foo'
    self.assertEqual((0, 127), tensor_spec_utils.bounds(tensor_spec))

  def test_max_less_than_min_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.max.int32s.array[:] = [-1]
    tensor_spec.min.int32s.array[:] = [1]
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*min 1.*max -1'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_min_legacy(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.min.uint32 = 1
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((1, 2**32 - 1), bounds)

  def test_max_legacy(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.UINT32
    tensor_spec.max.uint32 = 1
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((0, 1), bounds)

  def test_min_and_max_legacy(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT32
    tensor_spec.min.int32 = -1
    tensor_spec.max.int32 = 1
    bounds = tensor_spec_utils.bounds(tensor_spec)
    self.assertEqual((-1, 1), bounds)

  def test_max_larger_than_dtype_can_support_raises_error_legacy(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
    tensor_spec.max.int8 = 500
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*-128, 500.*int8.*-128, 127'):
      tensor_spec_utils.bounds(tensor_spec)

  def test_min_larger_than_dtype_can_support_raises_error(self):
    tensor_spec = dm_env_rpc_pb2.TensorSpec()
    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
    tensor_spec.min.int8 = -500
    tensor_spec.name = 'foo'
    with self.assertRaisesRegex(ValueError, 'foo.*-500, 127.*int8.*-128, 127'):
      tensor_spec_utils.bounds(tensor_spec)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""An implementation of a dm_env environment using dm_env_rpc."""

import dm_env

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import dm_env_utils
from dm_env_rpc.v1 import spec_manager

# Default observation names for common RL concepts.  By default the dm_env
# wrapper will use these for reward and discount if available, but this behavior
# can be overridden.
DEFAULT_REWARD_KEY = 'reward'
DEFAULT_DISCOUNT_KEY = 'discount'


class DmEnvAdaptor(dm_env.Environment):
  """An implementation of dm_env using dm_env_rpc as the data protocol."""

  def __init__(self, connection, specs, requested_observations=None):
    """Initializes the environment with the provided dm_env_rpc connection.

    Args:
      connection: An instance of Connection already connected to a dm_env_rpc
        server and after a successful JoinWorldRequest has been sent.
      specs: A dm_env_rpc ActionObservationSpecs message for the environment.
      requested_observations: The observation names to be requested from the
        environment when step is called. If None is specified then all
        observations will be requested.
    """
    self._dm_env_rpc_specs = specs
    self._action_specs = spec_manager.SpecManager(specs.actions)
    self._observation_specs = spec_manager.SpecManager(specs.observations)
    self._connection = connection
    self._last_state = dm_env_rpc_pb2.EnvironmentStateType.TERMINATED

    if requested_observations is None:
      requested_observations = self._observation_specs.names()
    requested_observations = set(requested_observations)

    self._is_reward_requested = DEFAULT_REWARD_KEY in requested_observations
    self._is_discount_requested = DEFAULT_DISCOUNT_KEY in requested_observations

    self._default_reward_spec = None
    self._default_discount_spec = None
    if DEFAULT_REWARD_KEY in self._observation_specs.names():
      self._default_reward_spec = dm_env_utils.tensor_spec_to_dm_env_spec(
          self._observation_specs.name_to_spec(DEFAULT_REWARD_KEY))
      requested_observations.add(DEFAULT_REWARD_KEY)
    if DEFAULT_DISCOUNT_KEY in self._observation_specs.names():
      self._default_discount_spec = (
          dm_env_utils.tensor_spec_to_dm_env_spec(
              self._observation_specs.name_to_spec(DEFAULT_DISCOUNT_KEY)))
      requested_observations.add(DEFAULT_DISCOUNT_KEY)

    unsupported_observations = requested_observations.difference(
        self._observation_specs.names())
    if unsupported_observations:
      raise ValueError('Unsupported observations requested: {}'.format(
          unsupported_observations))
    self._requested_observation_uids = [
        self._observation_specs.name_to_uid(name)
        for name in requested_observations
    ]

    # Not strictly necessary but it makes the unit tests deterministic.
    self._requested_observation_uids.sort()

  def reset(self):
    """Implements dm_env.Environment.reset."""
    response = self._connection.send(dm_env_rpc_pb2.ResetRequest())
    if self._dm_env_rpc_specs != response.specs:
      raise RuntimeError('Environment changed spec after reset')
    self._last_state = dm_env_rpc_pb2.EnvironmentStateType.INTERRUPTED
    return self.step({})

  def step(self, actions):
    """Implements dm_env.Environment.step."""
    step_response = self._connection.send(
        dm_env_rpc_pb2.StepRequest(
            requested_observations=self._requested_observation_uids,
            actions=self._action_specs.pack(actions)))

    observations = self._observation_specs.unpack(step_response.observations)

    if (step_response.state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING and
        self._last_state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING):
      step_type = dm_env.StepType.MID
    elif step_response.state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING:
      step_type = dm_env.StepType.FIRST
    elif self._last_state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING:
      step_type = dm_env.StepType.LAST
    else:
      raise RuntimeError('Environment transitioned from {} to {}'.format(
          self._last_state, step_response.state))

    self._last_state = step_response.state

    reward = self.reward(
        state=step_response.state,
        step_type=step_type,
        observations=observations)
    discount = self.discount(
        state=step_response.state,
        step_type=step_type,
        observations=observations)
    if not self._is_reward_requested:
      observations.pop(DEFAULT_REWARD_KEY, None)
    if not self._is_discount_requested:
      observations.pop(DEFAULT_DISCOUNT_KEY, None)
    return dm_env.TimeStep(step_type, reward, discount, observations)

  def reward(self, state, step_type, observations):
    """Returns the reward for the given observation state.

    Override in inherited classes to give different reward functions.

    Args:
      state: A dm_env_rpc EnvironmentStateType enum describing the state of the
        environment.
      step_type: The dm_env StepType describing the state of the environment.
      observations: The unpacked observations dictionary mapping string keys to
        scalars and NumPy arrays.

    Returns:
      A reward for the given step.  The shape and type matches that returned by
      `self.reward_spec()`.
    """
    if step_type == dm_env.StepType.FIRST:
      return None
    elif self._default_reward_spec:
      return observations[DEFAULT_REWARD_KEY]
    else:
      return 0.0

  def discount(self, state, step_type, observations):
    """Returns the discount for the given observation state.

    Override in inherited classes to give different discount functions.

    Args:
      state: A dm_env_rpc EnvironmentStateType enum describing the state of the
        environment.
      step_type: The dm_env StepType describing the state of the environment.
      observations: The unpacked observations dictionary mapping string keys to
        scalars and NumPy arrays.

    Returns:
      The discount for the given step.  The shape and type matches that returned
      by `self.discount_spec()`.
    """
    if self._default_discount_spec:
      return observations[DEFAULT_DISCOUNT_KEY]
    if step_type == dm_env.StepType.FIRST:
      return None
    elif (state == dm_env_rpc_pb2.EnvironmentStateType.RUNNING or
          state == dm_env_rpc_pb2.EnvironmentStateType.INTERRUPTED):
      return 1.0
    else:
      return 0.0

  def observation_spec(self):
    """Implements dm_env.Environment.observation_spec."""
    specs = {}
    for uid in self._requested_observation_uids:
      name = self._observation_specs.uid_to_name(uid)
      specs[name] = dm_env_utils.tensor_spec_to_dm_env_spec(
          self._observation_specs.uid_to_spec(uid))
    if not self._is_reward_requested:
      specs.pop(DEFAULT_REWARD_KEY, None)
    if not self._is_discount_requested:
      specs.pop(DEFAULT_DISCOUNT_KEY, None)
    return specs

  def action_spec(self):
    """Implements dm_env.Environment.action_spec."""
    return dm_env_utils.dm_env_spec(self._action_specs)

  def reward_spec(self):
    """Implements dm_env.Environment.reward_spec."""
    return (self._default_reward_spec or
            super(DmEnvAdaptor, self).reward_spec())

  def discount_spec(self):
    """Implements dm_env.Environment.discount_spec."""
    return (self._default_discount_spec or
            super(DmEnvAdaptor, self).discount_spec())

  def close(self):
    """Implements dm_env.Environment.close."""
    # Leaves the world if we were joined.  If not, this will be a no-op anyway.
    self._connection.send(dm_env_rpc_pb2.LeaveWorldRequest())
    self._connection = None



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Helper Python utilities for bridging dm_env_rpc and NumPy.

Note that the Tensor proto payload type is not supported, as it doesn't play
well with NumPy.
"""

import struct

import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2


class _BytesWrapper(object):
  """Allows protobuf bytes field to be set using the [:] = ... syntax."""

  def __init__(self, array, signed):
    self._array = array
    self._signed = signed

  def __getitem__(self, index):
    return self._to_scalar(self._array.array[index])

  def __setitem__(self, index, value):
    if index == slice(None, None, None):
      self._array.array = struct.pack(
          str(len(value)) + ('b' if self._signed else 'B'), *value)
    else:
      raise ValueError('Unsupported index {}'.format(index))

  def __len__(self):
    return len(self._array.array)

  def _to_scalar(self, value):
    return np.int8(value) if self._signed else np.uint8(value)

  def as_np_array(self):
    return np.frombuffer(
        self._array.array, dtype=np.int8 if self._signed else np.uint8)


# Payload channel name, NumPy representation, and the payload array.
_RAW_ASSOCIATIONS = (
    ('floats', np.float32, lambda tensor: tensor.floats.array),
    ('doubles', np.float64, lambda tensor: tensor.doubles.array),
    ('int8s', np.int8, lambda tensor: _BytesWrapper(tensor.int8s, signed=True)),
    ('int32s', np.int32, lambda tensor: tensor.int32s.array),
    ('int64s', np.int64, lambda tensor: tensor.int64s.array),
    ('uint8s', np.uint8,
     lambda tensor: _BytesWrapper(tensor.uint8s, signed=False)),
    ('uint32s', np.uint32, lambda tensor: tensor.uint32s.array),
    ('uint64s', np.uint64, lambda tensor: tensor.uint64s.array),
    ('bools', np.bool_, lambda tensor: tensor.bools.array),
    ('strings', np.str_, lambda tensor: tensor.strings.array),
)

_NAME_TO_NP_TYPE = {
    name: np_type for name, np_type, payload in _RAW_ASSOCIATIONS
}

_TYPE_TO_PAYLOAD = {
    np_type: payload for name, np_type, payload in _RAW_ASSOCIATIONS
}

_DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE = {
    dm_env_rpc_pb2.DataType.FLOAT: np.float32,
    dm_env_rpc_pb2.DataType.DOUBLE: np.float64,
    dm_env_rpc_pb2.DataType.INT8: np.int8,
    dm_env_rpc_pb2.DataType.INT32: np.int32,
    dm_env_rpc_pb2.DataType.INT64: np.int64,
    dm_env_rpc_pb2.DataType.UINT8: np.uint8,
    dm_env_rpc_pb2.DataType.UINT32: np.uint32,
    dm_env_rpc_pb2.DataType.UINT64: np.uint64,
    dm_env_rpc_pb2.DataType.BOOL: np.bool_,
    dm_env_rpc_pb2.DataType.STRING: np.str_,
}


def get_tensor_type(tensor_proto):
  """Returns the NumPy type for the given tensor."""
  payload = tensor_proto.WhichOneof('payload')
  np_type = _NAME_TO_NP_TYPE.get(payload)
  if not np_type:
    raise TypeError('Unknown type {}'.format(payload))
  return np_type


def data_type_to_np_type(dm_env_rpc_dtype):
  """Returns the NumPy type for the given dm_env_rpc DataType."""
  np_type = _DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE.get(dm_env_rpc_dtype)
  if not np_type:
    raise TypeError('Unknown type {}'.format(dm_env_rpc_dtype))
  return np_type


def unpack_proto(proto, shape):
  """Converts a proto with payload oneof to a scalar or NumPy array.

  Args:
    proto: A dm_env_rpc proto with payload oneof.
    shape: Optional dimensions of the payload data. If not set or empty, the
      data is assumed to be a scalar type.

  Returns:
    If `shape` is empty or None,, returns a scalar (float, int, string, etc.)
    of the correct type and value. Otherwise returns a NumPy array of the
    payload with the correct type and shape.
  """
  np_type = get_tensor_type(proto)
  payload = _TYPE_TO_PAYLOAD[np_type](proto)
  if shape:
    if len(payload) == 1:
      array = np.full(np.maximum(shape, 1), payload[0])
    else:
      if isinstance(payload, _BytesWrapper):
        array = payload.as_np_array()
      else:
        array = np.array(payload, np_type)
      array.shape = shape
    return array
  else:
    length = len(payload)
    if length != 1:
      raise ValueError(
          'Scalar tensors must have exactly 1 element but had {} elements.'
          .format(length))
    return np_type(payload[0])


def unpack_tensor(tensor_proto):
  """Converts a Tensor proto to a scalar or NumPy array.

  Args:
    tensor_proto: A dm_env_rpc Tensor protobuf.

  Returns:
    If the provided tensor_proto has a non-empty `shape` attribute, returns
    a NumPy array of the payload with the correct type and shape.  If the
    `shape` attribute is empty, returns a scalar (float, int, string, etc.)
    of the correct type and value.
  """
  return unpack_proto(tensor_proto, tensor_proto.shape)


def pack_tensor(value, dtype=None, try_compress=False):
  """Encodes the given value as a tensor.

  Args:
    value: A scalar (float, int, string, etc.), NumPy array, or nested lists.
    dtype: The type to pack the data to.  If set to None, will attempt to detect
      the correct type automatically.  Either a dm_env_rpc DataType enum or
      NumPy type is acceptable.
    try_compress: A bool, whether to try and encode the tensor in less space or
      not. This will increase the computational cost of the packing, but may
      reduce the on-the-wire size of the tensor.  There are no guarantees that
      any compression will actually happen.

  Raises:
    ValueError: If `value` is a jagged array, not a primitive type or nested
      iterable of primitive types, or all elements can't be cast to the same
      type or the requested type.

  Returns:
    A dm_env_rpc Tensor proto containing the data.
  """
  packed = dm_env_rpc_pb2.Tensor()
  value = np.asarray(value)
  if value.dtype == np.object:
    raise ValueError('Could not convert to a tensor of primitive types.  Are '
                     'the iterables jagged?  Or are the data types not '
                     'primitive scalar types like strings, floats, or '
                     'integers?')
  if dtype is not None:
    value = value.astype(
        dtype=_DM_ENV_RPC_DTYPE_TO_NUMPY_DTYPE.get(dtype, dtype),
        copy=False,
        casting='same_kind')
  packed.shape[:] = value.shape
  pack_target = _TYPE_TO_PAYLOAD[value.dtype.type](packed)
  if (try_compress and np.all(value == next(value.flat))):
    # All elements are the same.  Pack in to a single value.
    pack_target[:] = [next(value.flat)]
  else:
    pack_target[:] = np.ravel(value).tolist()
  return packed



# Lint as: python3
# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A compliance mixin for testing that a server complies with dm_env_rpc."""

import abc
import functools
import operator

from absl.testing import absltest
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import error
from dm_env_rpc.v1 import tensor_spec_utils
from dm_env_rpc.v1 import tensor_utils


def _find_uid_not_in_set(uid_set):
  """Finds an example UID not in `uid_set`."""
  uids = set(uid_set)
  uid = 0
  while uid in uids:
    uid = uid + 1
  return uid


def _is_numeric_type(dtype):
  return (dtype != dm_env_rpc_pb2.DataType.PROTO and
          np.issubdtype(tensor_utils.data_type_to_np_type(dtype), np.number))


def _assert_less_equal(x, y, err_msg='', verbose=True):
  np.testing.assert_array_compare(
      operator.__le__, x, y, err_msg=err_msg, verbose=verbose,
      header='Arrays are not less or equal ordered', equal_inf=False)


def _assert_greater_equal(x, y, err_msg='', verbose=True):
  np.testing.assert_array_compare(
      operator.__ge__, x, y, err_msg=err_msg, verbose=verbose,
      header='Arrays are not greater or equal ordered', equal_inf=False)


def _create_test_value(spec):
  """Creates a scalar test value consistent with the TensorSpec `spec`."""
  if _is_numeric_type(spec.dtype):
    return tensor_spec_utils.bounds(spec).min,
  else:
    np_type = tensor_utils.data_type_to_np_type(spec.dtype)
    return np_type()


def _create_test_tensor(spec, dtype=None):
  """Creates an arbitrary tensor consistent with the TensorSpec `spec`."""
  shape = np.asarray(spec.shape)
  shape[shape < 0] = 1
  value = [_create_test_value(spec)] * int(np.prod(shape))
  tensor = tensor_utils.pack_tensor(value, dtype=dtype or spec.dtype)
  tensor.shape[:] = shape
  return tensor


def _np_range_info(np_type):
  """Returns type info for `np_type`, which includes min and max attributes."""
  if issubclass(np_type, np.floating):
    return np.finfo(np_type)
  elif issubclass(np_type, np.integer):
    return np.iinfo(np_type)
  else:
    raise ValueError('{} does not have range info.'.format(np_type))


def _below_min(spec):
  """Returns a value below spec's min or None if none."""
  if not spec.HasField('min'):
    return None

  np_type = tensor_utils.data_type_to_np_type(spec.dtype)
  min_type_value = _np_range_info(np_type).min

  if min_type_value < tensor_spec_utils.bounds(spec).min:
    return min_type_value
  else:
    return None


def _above_max(spec):
  """Returns a value above spec's max or None if none."""
  if not spec.HasField('max'):
    return None

  np_type = tensor_utils.data_type_to_np_type(spec.dtype)
  max_type_value = _np_range_info(np_type).max

  if max_type_value > tensor_spec_utils.bounds(spec).max:
    return max_type_value
  else:
    return None


def _step_before_test(function):
  """Decorator which calls step before test function is run."""
  @functools.wraps(function)
  def wrapper(self, *args, **kwargs):
    # First step's actions are ignored, so step once to start the sequence.
    step_response = self.step()
    self.assertEqual(
        step_response.state, dm_env_rpc_pb2.EnvironmentStateType.RUNNING)
    return function(self, *args, **kwargs)
  return wrapper


class StepComplianceTestCase(absltest.TestCase, metaclass=abc.ABCMeta):
  """A base class for compliance tests."""

  @property
  @abc.abstractmethod
  def connection(self):
    """An instance of dm_env_rpc's Connection already joined to a world."""
    pass

  @property
  @abc.abstractmethod
  def specs(self):
    """The specs from a JoinWorldResponse."""
    pass

  @property
  def observation_uids(self):
    return set(self.specs.observations.keys())

  @property
  def action_uids(self):
    return set(self.specs.actions.keys())

  @property
  def numeric_actions(self):
    return {uid: spec for uid, spec in self.specs.actions.items()
            if _is_numeric_type(spec.dtype)}

  @property
  def nonnumeric_actions(self):
    return {uid: spec for uid, spec in self.specs.actions.items()
            if not _is_numeric_type(spec.dtype)}

  def step(self, **kwargs):
    """Sends a StepRequest and returns the StepResponse."""
    return self.connection.send(dm_env_rpc_pb2.StepRequest(**kwargs))

  # pylint: disable=missing-docstring
  ##############################################################################
  # Observations
  ##############################################################################
  def test_no_observations_returned_if_not_requested(self):
    observations = self.step().observations
    self.assertEqual({}, observations)

  def test_requested_observations_are_returned(self):
    response = self.step(requested_observations=self.observation_uids)
    observations = response.observations
    self.assertEqual(self.observation_uids, set(observations.keys()))

  def test_cannot_request_invalid_observation_uid(self):
    bad_uid = _find_uid_not_in_set(self.observation_uids)
    with self.assertRaisesRegex(error.DmEnvRpcError, str(bad_uid)):
      self.step(requested_observations=[bad_uid])

  def test_all_observation_dtypes_match_spec_dtypes(self):
    response = self.step(requested_observations=self.observation_uids)
    for uid, observation in response.observations.items():
      spec = self.specs.observations[uid]
      with self.subTest(uid=uid, name=spec.name):
        spec_type = tensor_utils.data_type_to_np_type(spec.dtype)
        tensor_type = tensor_utils.get_tensor_type(observation)
        self.assertEqual(spec_type, tensor_type)

  def test_all_numerical_observations_in_range(self):
    numeric_uids = (uid for uid, spec in self.specs.observations.items()
                    if _is_numeric_type(spec.dtype))
    response = self.step(requested_observations=numeric_uids)
    for uid, observation in response.observations.items():
      spec = self.specs.observations[uid]
      with self.subTest(uid=uid, name=spec.name):
        unpacked = tensor_utils.unpack_tensor(observation)
        bounds = tensor_spec_utils.bounds(spec)
        _assert_less_equal(unpacked, bounds.max)
        _assert_greater_equal(unpacked, bounds.min)

  def test_duplicated_requested_observations_are_redundant(self):
    response = self.step(requested_observations=list(self.observation_uids) * 2)
    self.assertEqual(self.observation_uids, set(response.observations.keys()))

  def test_can_request_each_observation_individually(self):
    for uid in self.observation_uids:
      spec = self.specs.observations[uid]
      with self.subTest(uid=uid, name=spec.name):
        response = self.step(requested_observations=[uid])
        self.assertEqual([uid], list(response.observations.keys()))

  ##############################################################################
  # Actions
  ##############################################################################
  def test_first_step_actions_are_ignored(self):
    bad_uid = _find_uid_not_in_set(self.action_uids)
    self.step(actions={bad_uid: tensor_utils.pack_tensor(0)})

  @_step_before_test
  def test_can_send_each_action_individually(self):
    for uid, spec in self.specs.actions.items():
      with self.subTest(uid=uid, name=spec.name):
        tensor = _create_test_tensor(spec)
        self.step(actions={uid: tensor})

  @_step_before_test
  def test_cannot_send_wrong_numeric_type_action(self):
    for uid, spec in self.numeric_actions.items():
      with self.subTest(uid=uid, name=spec.name):
        np_type = tensor_utils.data_type_to_np_type(spec.dtype)
        wrong_dtype = (np.int32 if np_type == np.issubdtype(np_type, np.inexact)
                       else np.float32)
        tensor = _create_test_tensor(spec, dtype=wrong_dtype)
        with self.assertRaises(error.DmEnvRpcError):
          self.step(actions={uid: tensor})

  @_step_before_test
  def test_cannot_send_wrong_type_to_nonnumeric_actions(self):
    tensor = tensor_utils.pack_tensor(0, dtype=np.int32)
    for uid, spec in self.nonnumeric_actions.items():
      with self.subTest(uid=uid, name=spec.name):
        shape = np.asarray(spec.shape)
        shape[shape < 0] = 1
        tensor.shape[:] = shape
        with self.assertRaises(error.DmEnvRpcError):
          self.step(actions={uid: tensor})

  @_step_before_test
  def test_cannot_send_invalid_action_uid(self):
    bad_uid = _find_uid_not_in_set(self.action_uids)
    with self.assertRaises(error.DmEnvRpcError):
      self.step(actions={bad_uid: tensor_utils.pack_tensor(0)})

  @_step_before_test
  def test_cannot_send_action_below_min(self):
    for uid, spec in self.numeric_actions.items():
      with self.subTest(uid=uid, name=spec.name):
        below = _below_min(spec)
        if below is None:
          # There are no values below spec's min.
          continue
        tensor = tensor_utils.pack_tensor(below, dtype=spec.dtype)
        shape = np.asarray(spec.shape)
        shape[shape < 0] = 1
        tensor.shape[:] = shape
        with self.assertRaises(error.DmEnvRpcError):
          self.step(actions={uid: tensor})

  @_step_before_test
  def test_cannot_send_action_above_max(self):
    for uid, spec in self.numeric_actions.items():
      with self.subTest(uid=uid, name=spec.name):
        above = _above_max(spec)
        if above is None:
          # There are no values above spec's max.
          continue
        tensor = tensor_utils.pack_tensor(above, dtype=spec.dtype)
        shape = np.asarray(spec.shape)
        shape[shape < 0] = 1
        tensor.shape[:] = shape
        with self.assertRaises(error.DmEnvRpcError):
          self.step(actions={uid: tensor})

  @_step_before_test
  def test_cannot_send_action_with_wrong_shape(self):
    for uid, spec in self.specs.actions.items():
      with self.subTest(uid=uid, name=spec.name):
        tensor = _create_test_tensor(spec)
        # Add too many dimensions to shape.
        tensor.shape[:] = tensor.shape[:] + [1]
        with self.assertRaises(error.DmEnvRpcError):
          self.step(actions={uid: tensor})

  @_step_before_test
  def test_can_send_variable_dimension_tensor_action(self):
    actions_with_shape = {uid: spec for uid, spec in self.specs.actions.items()
                          if spec.shape}
    for uid, spec in actions_with_shape.items():
      with self.subTest(uid=uid, name=spec.name):
        tensor = _create_test_tensor(spec)
        # Set first dimension to be variable.
        tensor.shape[0] = -1
        self.step(actions={uid: tensor})

  @_step_before_test
  def test_cannot_send_tensor_with_too_many_variable_dimensions(self):
    actions_with_multidimensional_shape = {
        uid: spec for uid, spec in self.specs.actions.items()
        if len(spec.shape) >= 2}
    for uid, spec in actions_with_multidimensional_shape.items():
      with self.subTest(uid=uid, name=spec.name):
        tensor = _create_test_tensor(spec)
        # Set multiple variable dimensions.
        tensor.shape[0] = -1
        tensor.shape[1] = -1
        with self.assertRaises(error.DmEnvRpcError):
          self.step(actions={uid: tensor})

  @_step_before_test
  def test_can_send_broadcastable_actions(self):
    for uid, spec in self.specs.actions.items():
      with self.subTest(uid=uid, name=spec.name):
        tensor = tensor_utils.pack_tensor(
            _create_test_value(spec), dtype=spec.dtype)
        shape = np.asarray(spec.shape)
        shape[shape < 0] = 1
        tensor.shape[:] = shape
        self.step(actions={uid: tensor})
  # pylint: enable=missing-docstring



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Helper Python utilities for managing dm_env_rpc TensorSpecs."""

import collections
import numpy as np

from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import tensor_utils


Bounds = collections.namedtuple('Bounds', ['min', 'max'])

_SCALAR_VALUE_TYPES = frozenset(
    ('float', 'double', 'int8', 'int32', 'int64', 'uint8', 'uint32', 'uint64'))


def _np_range_info(np_type):
  """Returns type info for `np_type`, which includes min and max attributes."""
  if issubclass(np_type, np.floating):
    return np.finfo(np_type)
  elif issubclass(np_type, np.integer):
    return np.iinfo(np_type)
  else:
    raise ValueError('{} does not have range info.'.format(np_type))


def _get_value(min_max_value, shape, default):
  """Helper function that returns the min/max bounds for a Value message.

  Args:
    min_max_value: Value protobuf message to get value from.
    shape: Optional dimensions to unpack payload data to.
    default: Value to use if min_max_value is not set.

  Returns:
    A scalar if `shape` is empty or None, or an unpacked NumPy array of either
    the unpacked value or provided default.

  """
  which = min_max_value.WhichOneof('payload')
  value = which and getattr(min_max_value, which)

  if value is None:
    min_max = np.broadcast_to(default, shape) if shape else default
  elif which in _SCALAR_VALUE_TYPES:
    min_max = np.broadcast_to(value, shape) if shape else value
  else:
    min_max = tensor_utils.unpack_proto(min_max_value, shape)

  if (shape is not None
      and np.any(np.array(shape) < 0)
      and np.asarray(min_max).size > 1):
    raise ValueError(
        "TensorSpec's with variable length shapes can only have scalar ranges. "
        'Shape: {}, value: {}'.format(shape, min_max))
  return min_max


def bounds(tensor_spec):
  """Gets the inclusive bounds of `tensor_spec`.

  Args:
    tensor_spec: An instance of a dm_env_rpc TensorSpec proto.

  Returns:
    A named tuple (`min`, `max`) of inclusive bounds.

  Raises:
    ValueError: `tensor_spec` does not have a numeric dtype, or the type of its
      `min` or `max` does not match its dtype, or the the bounds are invalid in
      some way.
  """
  np_type = tensor_utils.data_type_to_np_type(tensor_spec.dtype)
  tensor_spec_type = dm_env_rpc_pb2.DataType.Name(tensor_spec.dtype).lower()
  if not issubclass(np_type, np.number):
    raise ValueError('TensorSpec "{}" has non-numeric type {}.'
                     .format(tensor_spec.name, tensor_spec_type))

  min_which = tensor_spec.min.WhichOneof('payload')
  if min_which and not min_which.startswith(tensor_spec_type):
    raise ValueError('TensorSpec "{}" has dtype {} but min type {}.'.format(
        tensor_spec.name, tensor_spec_type, min_which))

  max_which = tensor_spec.max.WhichOneof('payload')
  if max_which and not max_which.startswith(tensor_spec_type):
    raise ValueError('TensorSpec "{}" has dtype {} but max type {}.'.format(
        tensor_spec.name, tensor_spec_type, max_which))

  dtype_bounds = _np_range_info(np_type)
  min_bound = _get_value(tensor_spec.min, tensor_spec.shape, dtype_bounds.min)
  max_bound = _get_value(tensor_spec.max, tensor_spec.shape, dtype_bounds.max)

  if (np.any(min_bound < dtype_bounds.min) or
      np.any(max_bound > dtype_bounds.max)):
    raise ValueError(
        'TensorSpec "{}"\'s bounds [{}, {}] are larger than the bounds on its '
        '{} dtype [{}, {}]'.format(
            tensor_spec.name, min_bound, max_bound, tensor_spec_type,
            dtype_bounds.min, dtype_bounds.max))

  if np.any(max_bound < min_bound):
    raise ValueError('TensorSpec "{}" has min {} larger than max {}.'.format(
        tensor_spec.name, min_bound, max_bound))

  return Bounds(np_type(min_bound), np_type(max_bound))



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Install script for setuptools."""

import importlib.machinery
import os

from distutils.cmd import Command
import pkg_resources
from setuptools import find_packages
from setuptools import setup
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
GOOGLE_COMMON_PROTOS_ROOT_DIR = os.path.join(ROOT_DIR,
                                             'third_party/api-common-protos')


class _GenerateProtoFiles(Command):
  """Command to generate protobuf bindings for dm_env_rpc.proto."""

  descriptions = 'Generates Python protobuf bindings for dm_env_rpc.proto.'
  user_options = []

  def initialize_options(self):
    pass

  def finalize_options(self):
    pass

  def run(self):
    # Import grpc_tools here, after setuptools has installed setup_requires
    # dependencies.
    from grpc_tools import protoc  # pylint: disable=g-import-not-at-top

    if not os.path.exists(
        os.path.join(GOOGLE_COMMON_PROTOS_ROOT_DIR, 'google/rpc/status.proto')):
      raise RuntimeError(
          'Cannot find third_party/api-common-protos. '
          'Please run `git submodule init && git submodule update` to install '
          'the api-common-protos submodule.'
      )
    dm_env_rpc_proto = os.path.join(ROOT_DIR, 'dm_env_rpc/v1/dm_env_rpc.proto')
    grpc_protos_include = pkg_resources.resource_filename(
        'grpc_tools', '_proto')
    proto_args = [
        'grpc_tools.protoc',
        '--proto_path={}'.format(GOOGLE_COMMON_PROTOS_ROOT_DIR),
        '--proto_path={}'.format(grpc_protos_include),
        '--proto_path={}'.format(ROOT_DIR),
        '--python_out={}'.format(ROOT_DIR),
        '--grpc_python_out={}'.format(ROOT_DIR),
        dm_env_rpc_proto,
    ]
    if protoc.main(proto_args) != 0:
      raise RuntimeError('ERROR: {}'.format(proto_args))


class _BuildExt(build_ext):
  """Generate protobuf bindings in build_ext stage."""

  def run(self):
    self.run_command('generate_protos')
    build_ext.run(self)


class _BuildPy(build_py):
  """Generate protobuf bindings in build_py stage."""

  def run(self):
    self.run_command('generate_protos')
    build_py.run(self)


setup(
    name='dm-env-rpc',
    version=importlib.machinery.SourceFileLoader(
        '_version', 'dm_env_rpc/_version.py').load_module().__version__,
    description='A networking protocol for agent-environment communication.',
    author='DeepMind',
    license='Apache License, Version 2.0',
    keywords='reinforcement-learning python machine learning',
    packages=find_packages(exclude=['examples']),
    install_requires=[
        'dm-env>=1.2',
        'googleapis-common-protos',
        'grpcio',
        'numpy',
        'protobuf>=3.8',
    ],
    tests_require=[
        'absl-py',
        'nose',
        'mock',
        'portpicker',
    ],
    python_requires='>=3.5',
    setup_requires=['grpcio-tools'],
    extras_require={
        'examples': ['pygame', 'portpicker'],
    },
    cmdclass={
        'build_ext': _BuildExt,
        'build_py': _BuildPy,
        'generate_protos': _GenerateProtoFiles,
    },
    test_suite='nose.collector',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for dm_env_rpc/dm_env adaptor."""

from absl.testing import absltest
import dm_env
from dm_env import specs
import mock
import numpy as np

from dm_env_rpc.v1 import dm_env_adaptor
from dm_env_rpc.v1 import dm_env_rpc_pb2
from dm_env_rpc.v1 import tensor_utils

_SAMPLE_STEP_REQUEST = dm_env_rpc_pb2.StepRequest(
    requested_observations=[1, 2],
    actions={
        1: tensor_utils.pack_tensor(4, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('hello')
    })
_SAMPLE_STEP_RESPONSE = dm_env_rpc_pb2.StepResponse(
    state=dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
    observations={
        1: tensor_utils.pack_tensor(5, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('goodbye')
    })
_TERMINATED_STEP_RESPONSE = dm_env_rpc_pb2.StepResponse(
    state=dm_env_rpc_pb2.EnvironmentStateType.TERMINATED,
    observations={
        1: tensor_utils.pack_tensor(5, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('goodbye')
    })
_SAMPLE_SPEC = dm_env_rpc_pb2.ActionObservationSpecs(
    actions={
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo'),
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar')
    },
    observations={
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo'),
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar')
    },
)
_SAMPLE_SPEC_REORDERED = dm_env_rpc_pb2.ActionObservationSpecs(
    observations={
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar'),
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo')
    },
    actions={
        2: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.STRING, name='bar'),
        1: dm_env_rpc_pb2.TensorSpec(dtype=dm_env_rpc_pb2.UINT8, name='foo')
    },
)
# Ensures the equality check in reset() works if the dictionary elements are
# created in a different order.
_SAMPLE_RESET_RESPONSE = dm_env_rpc_pb2.ResetResponse(
    specs=_SAMPLE_SPEC_REORDERED)
_RESET_CHANGES_SPEC_RESPONSE = dm_env_rpc_pb2.ResetResponse(
    specs=dm_env_rpc_pb2.ActionObservationSpecs())

_RESERVED_SPEC = dm_env_rpc_pb2.ActionObservationSpecs(
    actions={},
    observations={
        1:
            dm_env_rpc_pb2.TensorSpec(
                dtype=dm_env_rpc_pb2.UINT8,
                name=dm_env_adaptor.DEFAULT_REWARD_KEY),
        2:
            dm_env_rpc_pb2.TensorSpec(
                dtype=dm_env_rpc_pb2.STRING,
                name=dm_env_adaptor.DEFAULT_DISCOUNT_KEY)
    })
_RESERVED_STEP_RESPONSE = dm_env_rpc_pb2.StepResponse(
    state=dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
    observations={
        1: tensor_utils.pack_tensor(5, dtype=dm_env_rpc_pb2.UINT8),
        2: tensor_utils.pack_tensor('goodbye')
    })


class DmEnvAdaptorTests(absltest.TestCase):

  def setUp(self):
    super(DmEnvAdaptorTests, self).setUp()
    self._connection = mock.MagicMock()
    self._env = dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC)

  def test_requested_observations(self):
    requested_observations = ['foo']
    filtered_env = dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC,
                                               requested_observations)

    expected_filtered_step_request = dm_env_rpc_pb2.StepRequest(
        requested_observations=[1],
        actions={
            1: tensor_utils.pack_tensor(4, dtype=dm_env_rpc_pb2.UINT8),
            2: tensor_utils.pack_tensor('hello')
        })

    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    filtered_env.step({'foo': 4, 'bar': 'hello'})

    self._connection.send.assert_called_once_with(
        expected_filtered_step_request)

  def test_invalid_requested_observations(self):
    requested_observations = ['invalid']
    with self.assertRaisesRegex(ValueError,
                                'Unsupported observations requested'):
      dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC,
                                  requested_observations)

  def test_requested_observation_spec(self):
    requested_observations = ['foo']
    filtered_env = dm_env_adaptor.DmEnvAdaptor(self._connection, _SAMPLE_SPEC,
                                               requested_observations)
    observation_names = [name for name in filtered_env.observation_spec()]
    self.assertEqual(requested_observations, observation_names)

  def test_first_running_step(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    timestep = self._env.step({'foo': 4, 'bar': 'hello'})

    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self.assertEqual(dm_env.StepType.FIRST, timestep.step_type)
    self.assertEqual(None, timestep.reward)
    self.assertEqual(None, timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_mid_running_step(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    timestep = self._env.step({'foo': 4, 'bar': 'hello'})

    self.assertEqual(dm_env.StepType.MID, timestep.step_type)
    self.assertEqual(0.0, timestep.reward)
    self.assertEqual(1.0, timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_last_step(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self._connection.send = mock.MagicMock(
        return_value=_TERMINATED_STEP_RESPONSE)
    timestep = self._env.step({'foo': 4, 'bar': 'hello'})

    self.assertEqual(dm_env.StepType.LAST, timestep.step_type)
    self.assertEqual(0.0, timestep.reward)
    self.assertEqual(0.0, timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_illegal_state_transition(self):
    self._connection.send = mock.MagicMock(
        return_value=_TERMINATED_STEP_RESPONSE)
    with self.assertRaisesRegex(RuntimeError, 'Environment transitioned'):
      self._env.step({})

  def test_reset(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self._connection.send = mock.MagicMock(
        side_effect=[_SAMPLE_RESET_RESPONSE, _SAMPLE_STEP_RESPONSE])
    timestep = self._env.reset()

    self.assertEqual(dm_env.StepType.FIRST, timestep.step_type)
    self.assertEqual(None, timestep.reward)
    self.assertEqual(None, timestep.discount)
    self.assertEqual({'foo': 5, 'bar': 'goodbye'}, timestep.observation)

  def test_reset_changes_spec_raises_error(self):
    self._connection.send = mock.MagicMock(return_value=_SAMPLE_STEP_RESPONSE)
    self._env.step({'foo': 4, 'bar': 'hello'})
    self._connection.send.assert_called_once_with(_SAMPLE_STEP_REQUEST)
    self._connection.send = mock.MagicMock(
        side_effect=[_RESET_CHANGES_SPEC_RESPONSE, _SAMPLE_STEP_RESPONSE])
    with self.assertRaisesRegex(RuntimeError, 'changed spec'):
      self._env.reset()

  def test_observation_spec(self):
    expected_spec = {
        'foo': specs.Array(shape=(), dtype=np.uint8, name='foo'),
        'bar': specs.Array(shape=(), dtype=np.str_, name='bar')
    }
    self.assertEqual(expected_spec, self._env.observation_spec())

  def test_action_spec(self):
    expected_spec = {
        'foo': specs.Array(shape=(), dtype=np.uint8, name='foo'),
        'bar': specs.Array(shape=(), dtype=np.str_, name='bar')
    }
    self.assertEqual(expected_spec, self._env.action_spec())

  def test_cant_step_after_close(self):
    self._connection.send = mock.MagicMock(
        return_value=dm_env_rpc_pb2.LeaveWorldResponse())
    self._env.close()
    with self.assertRaisesRegex(AttributeError, 'send'):
      self._env.step({})

  def test_reward_spec_default(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.float64), self._env.reward_spec())

  def test_discount_spec_default(self):
    self.assertEqual(
        specs.BoundedArray(
            shape=(), dtype=np.float64, minimum=0.0, maximum=1.0),
        self._env.discount_spec())

  def test_close_leaves_world(self):
    self._connection.send = mock.MagicMock(
        return_value=dm_env_rpc_pb2.LeaveWorldResponse())
    self._env.close()
    self._connection.send.assert_called_once_with(
        dm_env_rpc_pb2.LeaveWorldRequest())

  def test_close_errors_when_cannot_leave_world(self):
    self._connection.send = mock.MagicMock(side_effect=ValueError('foo'))
    with self.assertRaisesRegex(ValueError, 'foo'):
      self._env.close()


class OverrideRewardDiscount(dm_env_adaptor.DmEnvAdaptor):

  def __init__(self):
    self.connection = mock.MagicMock()
    self.reward = mock.MagicMock()
    self.discount = mock.MagicMock()
    super(OverrideRewardDiscount, self).__init__(self.connection, _SAMPLE_SPEC)


class RewardDiscountOverrideTests(absltest.TestCase):

  def test_override_reward(self):
    env = OverrideRewardDiscount()
    env.reward.return_value = 0.5
    env.connection.send.return_value = _SAMPLE_STEP_RESPONSE
    timestep = env.step({})
    self.assertEqual(0.5, timestep.reward)
    env.reward.assert_called()
    self.assertEqual(dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
                     env.reward.call_args[1]['state'])
    self.assertEqual(dm_env.StepType.FIRST,
                     env.reward.call_args[1]['step_type'])
    self.assertDictEqual({
        'foo': 5,
        'bar': 'goodbye'
    }, env.reward.call_args[1]['observations'])

  def test_override_discount(self):
    env = OverrideRewardDiscount()
    env.discount.return_value = 0.5
    env.connection.send.return_value = _SAMPLE_STEP_RESPONSE
    timestep = env.step({})
    self.assertEqual(0.5, timestep.discount)
    env.discount.assert_called()
    self.assertEqual(dm_env_rpc_pb2.EnvironmentStateType.RUNNING,
                     env.discount.call_args[1]['state'])
    self.assertEqual(dm_env.StepType.FIRST,
                     env.discount.call_args[1]['step_type'])
    self.assertDictEqual({
        'foo': 5,
        'bar': 'goodbye'
    }, env.discount.call_args[1]['observations'])


class ReservedKeywordTests(absltest.TestCase):

  def setUp(self):
    super(ReservedKeywordTests, self).setUp()
    self._connection = mock.MagicMock()
    self._env = dm_env_adaptor.DmEnvAdaptor(self._connection, _RESERVED_SPEC)

  def test_reward_spec(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.uint8), self._env.reward_spec())

  def test_discount_spec(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.str_), self._env.discount_spec())

  def test_reward_from_reserved_keyword(self):
    self._connection.send = mock.MagicMock(return_value=_RESERVED_STEP_RESPONSE)
    self._env.step({})  # Reward is None for first step.
    timestep = self._env.step({})

    self.assertEqual(5, timestep.reward)

  def test_discount(self):
    self._connection.send = mock.MagicMock(return_value=_RESERVED_STEP_RESPONSE)
    timestep = self._env.step({})

    self.assertEqual('goodbye', timestep.discount)


class EnvironmentAutomaticallyRequestsReservedKeywords(absltest.TestCase):

  def setUp(self):
    super(EnvironmentAutomaticallyRequestsReservedKeywords, self).setUp()
    self._connection = mock.MagicMock()
    self._env = dm_env_adaptor.DmEnvAdaptor(
        self._connection, _RESERVED_SPEC, requested_observations=[])
    self._connection.send = mock.MagicMock(return_value=_RESERVED_STEP_RESPONSE)

  def test_reward_spec_unrequested(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.uint8), self._env.reward_spec())

  def test_discount_spec_unrequested(self):
    self.assertEqual(
        specs.Array(shape=(), dtype=np.str_), self._env.discount_spec())

  def test_does_not_give_back_unrequested_observations(self):
    timestep = self._env.step({})
    self.assertEqual({}, timestep.observation)

  def test_first_reward_none(self):
    timestep = self._env.step({})
    self.assertIsNone(timestep.reward)

  def test_reward_piped_correctly(self):
    self._env.step({})  # Reward is None for first step.
    timestep = self._env.step({})
    self.assertEqual(5, timestep.reward)

  def test_discount_piped_correctly(self):
    timestep = self._env.step({})
    self.assertEqual('goodbye', timestep.discount)


if __name__ == '__main__':
  absltest.main()
